"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""


# -*- coding: utf-8 -*-
import citrusleaf as cl
from helpers import convertor
from helpers import common
from helpers.debugger import Debugger
# from helpers.caching import CachedMethod
from constants import *
from helpers.info_dict import InfoDict
from datetime import datetime
from copy import deepcopy
from distutils import version

#from flaskapp import app
#logger = app.logger

debugger = Debugger()


@debugger.check_and_log
# @CachedMethod()
def is_num(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

@debugger.check_and_log
# @CachedMethod()
def to_num(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


@debugger.check_and_log
# @CachedMethod()
def get_node_id(ip, port, user=None, password=None, sock=None):

    return get_node_info(ip, port, user=user, password=password, sock=sock).get("node", None)

@debugger.check_and_log
def is_node_available(ip, port, user=None, password=None, sock=None):

    response = get_node_info(ip, port, user=user, password=password, sock=sock)
    if response == {} or response == -1:
        return False
    return True

@debugger.check_and_log
def get_node_build(ip, port, user=None, password=None, sock=None):

    node_build_id = get_node_info(ip, port, user=user, password=password, sock=sock).get("build", None)
    return node_build_id

@debugger.check_and_log
# @CachedMethod()
def is_xdr_enabled(ip, xdr_port, user=None, password=None, sock=None):

    if user is None:
        stats = cl.citrusleaf_info(ip, xdr_port, "statistics", sock=sock)
    else:
        stats = cl.citrusleaf_info(ip, xdr_port, "statistics", user=user, password=password, sock=sock)
    if stats in cl.ERROR_CODES:
        return False
    return True

@debugger.check_and_log
# @CachedMethod(60)
def get_self_address(ip, port, user=None, password=None, sock=None):

    if user is not None:
        node_address = cl.citrusleaf_info(ip, port, "service", user=user, password=password, sock=sock)
    else:
        node_address = cl.citrusleaf_info(ip, port, "service", sock=sock)
    if node_address == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    if node_address in cl.ERROR_CODES:
        return node_address
    return node_address.split(";")[0]  # multiple_interfaces_on_node

@debugger.check_and_log
def get_cluster_nodes(ip, port, user=None, password=None, sock=None):

    services = get_node_info(ip, port, user=user, password=password, sock=sock).get("services", UNABLE_TO_CONNECT)
    service = get_self_address(ip, port, user=user, password=password, sock=sock)
    if services == UNABLE_TO_CONNECT or service in cl.ERROR_CODES:
        return []
    nodes = convertor.text_to_list(services)
    nodes.append(service.split(";")[0])
    # for node_address in nodes:
    #     ip, port = node_address.split(":")
    #     node_id = get_node_id(ip, port, user, password)
    #     if node_id is not None and node_id not in node_list.keys():
    #         node_list[node_id] = node_address

    return nodes


@debugger.check_and_log
def get_node_statistics(ip, port, user=None, password=None, sock=None):

    # statistics = get_node_info(ip, port, user=user, password=password, clear_cache=True, sock=sock).get("statistics", None)
    statistics = get_node_info(ip, port, user=user, password=password, sock=sock).get("statistics", None)
    if statistics is None:
        return InfoDict(node_status="off")
    statistics = convertor.text_to_list(statistics)
    statistics = convertor.list_to_dict(statistics)
    statistics['timestamp'] = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    return statistics

@debugger.check_and_log
# @CachedMethod(300)
def get_node_info(ip, port, user=None, password=None, sock=None):

    if user is None:
        info = cl.citrusleaf_info(ip, port, sock=sock)
    else:
        info = cl.citrusleaf_info(ip, port, user=user, password=password, sock=sock)
    if info in cl.ERROR_CODES or info == {'security error - not authenticated': ''}:
        return {}
    return info

@debugger.check_and_log
# @CachedMethod()
def get_node_configuration(ip, port, user=None, password=None, sock=None):

    if user is None:
        configuration = cl.citrusleaf_info(ip, port, "get-config:", sock=sock)
    else:
        configuration = cl.citrusleaf_info(ip, port, "get-config:", user=user, password=password, sock=sock)
    if configuration in cl.ERROR_CODES:
        return InfoDict(node_status="off")
    configuration = convertor.text_to_list(configuration)
    return convertor.list_to_dict(configuration)

@debugger.check_and_log
def set_node_configuration(ip, port, parameter, value, user=None, password=None, sock=None):

    arg = "set-config:context=service;%s=%s" % (parameter, value)
    if user is None:
        if cl.citrusleaf_info(ip, port, arg, sock=sock) == 'ok':
            return True
    else:
        if cl.citrusleaf_info(ip, port, arg, user=user, password=password, sock=sock) == 'ok':
            return True
    return False

@debugger.check_and_log
def get_xdr_statistics(ip, port, user=None, password=None, sock=None):

    statistics = get_xdr_info(ip, port, user=user, password=password, sock=sock)
    if statistics is None:
        return InfoDict(node_status="off")
    statistics = convertor.text_to_list(statistics)
    statistics = convertor.list_to_dict(statistics)
    statistics['timestamp'] = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    return statistics

@debugger.check_and_log
# @CachedMethod()
def get_xdr_info(ip, port, user=None, password=None, sock=None):

    arg = "statistics/xdr"
    if user is None:
        info = cl.citrusleaf_info(ip, port, arg, sock=sock)
    else:
        info = cl.citrusleaf_info(ip, port, arg, user=user, password=password, sock=sock)
    if info in cl.ERROR_CODES:
        return None
    return info

@debugger.check_and_log
def set_xdr_configuration(ip, port, parameter, value, user=None, password=None, sock=None):

    arg = "set-config:context=xdr;%s=%s" % (parameter, value)
    if user is None:
        if cl.citrusleaf_info(ip, port, arg, sock=sock) == 'ok':
            return True
    else:
        if cl.citrusleaf_info(ip, port, arg, user=user, password=password, sock=sock) == 'ok':
            return True
    return False

@debugger.check_and_log
# @CachedMethod()
def get_node_bins_info(ip, port, user=None, password=None, sock=None):

    bins_info = dict()
    if user is None:
        temp = cl.citrusleaf_info(ip, port, "bins", sock=sock)
    else:
        temp = cl.citrusleaf_info(ip, port, "bins", user=user, password=password, sock=sock)
    if temp in cl.ERROR_CODES:
        return None

    for row in convertor.text_to_list(temp):
        if row:
            namespace, bin_info = convertor.text_to_list(row, ":")
            temp = convertor.text_to_list(bin_info, ",")
            bins_info[namespace] = convertor.list_to_dict(temp[:2])
            bins_info[namespace]["bins"] = temp[2:]

    return bins_info

@debugger.check_and_log
def configure_node_latency(ip, port, user=None, password=None, sock=None):

    if user is None:
        result = cl.citrusleaf_info(ip, port, "latency:", sock=sock)
    else:
        result = cl.citrusleaf_info(ip, port, "latency:", user=user, password=password, sock=sock)
    if result not in cl.ERROR_CODES and "error" in str(result):
        if user is None:
            if cl.citrusleaf_info(ip, port, "hist-track-start:back=3600;slice=10;", sock=sock) == "ok":
                return True
        else:
            if cl.citrusleaf_info(ip, port, "hist-track-start:back=3600;slice=10;", user=user, password=password, sock=sock) == "ok":
                return True
    return False

@debugger.check_and_log
# @CachedMethod()
def get_node_latency(ip, port, user=None, password=None, sock=None, build=None):
    if build is None or version.LooseVersion(build) < "3.9.0":
        return get_node_latency_before39(ip, port, user=user, password=password, sock=sock)
    else:
        return get_node_latency_after39(ip, port, user=user, password=password, sock=sock)


@debugger.check_and_log
# @CachedMethod()
def get_node_latency_before39(ip, port, user=None, password=None, sock=None):

    latency = dict()
    if user is None:
        result = cl.citrusleaf_info(ip, port, "latency:", sock=sock)
    else:
        result = cl.citrusleaf_info(ip, port, "latency:", user=user, password=password, sock=sock)

    if result in cl.ERROR_CODES or "error" in result:
        return {}

    rows = convertor.text_to_list(result)
    for i in range(0, len(rows), 2):

        if not rows[i]:
            continue
        ind = rows[i].index(':')
        op = rows[i][:ind]
        if op == "writes_reply":
            continue
        rows[i] = rows[i][ind + 1:]

        keys = convertor.text_to_list(rows[i], ",")
        values = convertor.text_to_list(rows[i + 1], ",")

        keys[0] = keys[0].split("-")[0]
        ops_per_sec = float(values[1])

        latency[op] = {'timestamp': common.time_average(values[0], keys[0]),
            'data': [],
            'ops/sec': ops_per_sec
            }

        less_than_1 = 0.0 if ops_per_sec == 0 else 100.0

        previous_key = ""
        previous_value = 0.0
        for key, value in zip(list(reversed(keys))[:len(keys)-2], list(reversed(values))[:len(keys)-2]):
            pct = float(value) - previous_value
            previous_value = float(value)
            original_key = key
            if previous_key != "":
                key = ">" + key[1:] + " to &#x2264;" + previous_key[1:]
            previous_key = original_key
            less_than_1 -= pct
            value = ops_per_sec * pct / 100
            latency[op]['data'] = [{key: dict(value=value, pct=pct)}] + latency[op]['data']

        less_than_1_value = ops_per_sec * less_than_1 / 100
        latency[op]['data'] = [{"&#x2264;1ms": dict(value=less_than_1_value, pct=less_than_1)}] + latency[op]['data']
    return latency

@debugger.check_and_log
def update_hist_pct(agg):
    ops = agg['ops/sec']
    if ops <= 0:
        return agg

    for idx, data in enumerate(agg["data"]):
        for k, v in data.iteritems():
            agg["data"][idx][k]['pct'] = agg["data"][idx][k]['value'] / ops * 100.0

    return agg

@debugger.check_and_log
def merge_hist(agg, d):
    if agg is None:
        return d
    else:
        if agg['timestamp'] < agg['timestamp']:
            agg['timestamp'] = agg['timestamp']

        agg['ops/sec'] += d['ops/sec']
        for idx, data in enumerate(agg["data"]):
            for k, v in data.iteritems():
                for nk, nv in d['data'][idx].iteritems():
                    agg["data"][idx][k]['value'] += nv['value']
        return agg


@debugger.check_and_log
# @CachedMethod()
def get_node_latency_after39(ip, port, user=None, password=None, sock=None):

    latency = dict()
    if user is None:
        result = cl.citrusleaf_info(ip, port, "latency:", sock=sock)
    else:
        result = cl.citrusleaf_info(ip, port, "latency:", user=user, password=password, sock=sock)

    if result in cl.ERROR_CODES:
        return {}

    rows = convertor.text_to_list(result)
    # remove err strings from the results
    rows = [x for x in rows if "error" not in x]
    for i in range(0, len(rows), 2):
        if not rows[i]:
            continue
        ind = rows[i].index(':')
        op = rows[i][:ind]
        if op == "writes_reply":
            continue
        rows[i] = rows[i][ind + 1:]

        keys = convertor.text_to_list(rows[i], ",")
        values = convertor.text_to_list(rows[i + 1], ",")

        keys[0] = keys[0].split("-")[0]
        ops_per_sec = float(values[1])

        latency[op] = {'timestamp': common.time_average(values[0], keys[0]),
            'data': [],
            'ops/sec': ops_per_sec
            }

        less_than_1 = 0.0 if ops_per_sec == 0 else 100.0

        previous_key = ""
        previous_value = 0.0
        for key, value in zip(list(reversed(keys))[:len(keys)-2], list(reversed(values))[:len(keys)-2]):
            pct = float(value) - previous_value
            previous_value = float(value)
            original_key = key
            if previous_key != "":
                key = ">" + key[1:] + " to &#x2264;" + previous_key[1:]
            previous_key = original_key
            less_than_1 -= pct
            value = ops_per_sec * pct / 100
            latency[op]['data'] = [{key: dict(value=value, pct=pct)}] + latency[op]['data']

        less_than_1_value = ops_per_sec * less_than_1 / 100
        latency[op]['data'] = [{"&#x2264;1ms": dict(value=less_than_1_value, pct=less_than_1)}] + latency[op]['data']

    reads = None
    writes = None
    writes_master = None
    query = None
    proxy = None
    udf = None
    # aggregate per namespace data into node data
    for k, v in latency.iteritems():
        if k.endswith("-read"):
            reads = merge_hist(reads, v)
        elif k.endswith("-write"):
            writes = merge_hist(writes, v)
        elif k.endswith("-query"):
            query = merge_hist(query, v)
        elif k.endswith("-proxy"):
            proxy = merge_hist(proxy, v)
        elif k.endswith("-udf"):
            udf = merge_hist(udf, v)

    latency = dict()
    if reads is not None:
        latency['reads'] = update_hist_pct(reads)
    if writes is not None:
        latency['writes'] = update_hist_pct(writes)
    if writes_master is not None:
        latency['writes_master'] = update_hist_pct(writes_master)
    if query is not None:
        latency['query'] = update_hist_pct(query)
    if proxy is not None:
        latency['proxy'] = update_hist_pct(proxy)
    if udf is not None:
        latency['udf'] = update_hist_pct(udf)

    return latency

@debugger.check_and_log
# @CachedMethod()
def get_namespaces(ip, port, user=None, password=None, sock=None):

    if user is None:
        namespaces = cl.citrusleaf_info(ip, port, "namespaces", sock=sock)
    else:
        namespaces = cl.citrusleaf_info(ip, port, "namespaces", user=user, password=password, sock=sock)
    if namespaces in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT
    return convertor.text_to_set(namespaces)

@debugger.check_and_log
# @CachedMethod()
def get_namespace_statistics(namespace, ip, port, user=None, password=None, sock=None):

    if user is None:
        stats = cl.citrusleaf_info(ip, port, "namespace/%s" % namespace, sock=sock)
    else:
        stats = cl.citrusleaf_info(ip, port, "namespace/%s" % namespace, user=user, password=password, sock=sock)
    if stats in cl.ERROR_CODES:
        return InfoDict(node_status="off")
    stats = convertor.text_to_list(stats)
    return convertor.list_to_dict(stats)

@debugger.check_and_log
def set_namespace_configuration(namespace, ip, port, parameter, value, user=None, password=None, sock=None):

    arg = "set-config:context=namespace;id=%s;%s=%s" % (namespace, parameter, value)
    if user is None:
        if cl.citrusleaf_info(ip, port, arg, sock=sock) == 'ok':
            return True
    else:
        if cl.citrusleaf_info(ip, port, arg, user=user, password=password, sock=sock) == 'ok':
            return True
    return False

@debugger.check_and_log
# @CachedMethod()
def get_sets_info(ip, port, user=None, password=None, sock=None):

    sets_info = dict()
    if user is None:
        result = cl.citrusleaf_info(ip, port, "sets", sock=sock)
    else:
        result = cl.citrusleaf_info(ip, port, "sets", user=user, password=password, sock=sock)

    if result in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT

    if result:
        for row in convertor.text_to_list(result.strip(), ";"):
            if not row:
                continue

            details = convertor.list_to_dict(
                convertor.text_to_list(row, ":"), "=")
            for key in list(details.keys()):  # For Aerospike 3.0.22+
                if key.startswith('set-'):
                    new_key = key.replace('set-', '')
                    details[new_key] = details.pop(key)

            if details.get('ns_name', None) is not None:
                namespace = sets_info.get(details['ns_name'], [])
                namespace.append(details)
                sets_info[details['ns_name']] = namespace

    return sets_info


@debugger.check_and_log
# @CachedMethod()
def get_agg_namespace_statistics(ip, port, user=None, password=None, sock=None):

    agg_stats = InfoDict()
    ns_list = get_namespaces(ip, port, user=user, password=password, sock=sock)
    if ns_list == UNABLE_TO_CONNECT:
        return None
    for ns in ns_list:
        stats = get_namespace_statistics(ns, ip, port, user=user, password=password, sock=sock)
        for stat, val in stats.iteritems():
            if is_num(val):
                agg_stats[stat] = agg_stats.get(stat, 0) + to_num(val)

    for stat, val in agg_stats.iteritems():
        agg_stats[stat] = str(val)

    return agg_stats

@debugger.check_and_log
# @CachedMethod()
def get_dc_statistics(ip, port, dc, user=None, password=None, sock=None):

    if user is None:
        statistics = cl.citrusleaf_info(ip, port, "dc/%s" % dc, sock=sock)
    else:
        statistics = cl.citrusleaf_info(ip, port, "dc/%s" % dc, user=user, password=password, sock=sock)
    if statistics in cl.ERROR_CODES:
        return InfoDict()
    statistics = convertor.text_to_list(statistics)
    return convertor.list_to_dict(statistics)

@debugger.check_and_log
# @CachedMethod()
def get_dc_names(ip, port, user=None, password=None, sock=None):

    if user is None:
        names = cl.citrusleaf_info(ip, port, "dcs", sock=sock)
    else:
        names = cl.citrusleaf_info(ip, port, "dcs", user=user, password=password, sock=sock)
    if names in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT

    return convertor.text_to_list(names)

@debugger.check_and_log
# @CachedMethod()
def get_udf_list(ip, port, user=None, password=None, sock=None):

    if user is None:
        lst = cl.citrusleaf_info(ip, port, "udf-list", sock=sock)
    else:
        lst = cl.citrusleaf_info(ip, port, "udf-list", user=user, password=password, sock=sock)
    if lst in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT

    udfs = []
    for record in convertor.text_to_list(lst, ";"):
        if record != '':
            udfs.append(convertor.list_to_dict(
                convertor.text_to_list(record, ",")))

    return udfs

@debugger.check_and_log
# @CachedMethod()
def get_scan_jobs(ip, port, user=None, password=None, sock=None):

    if user is None:
        lst = cl.citrusleaf_info(ip, port, "scan-list", sock=sock)
    else:
        lst = cl.citrusleaf_info(ip, port, "scan-list", user=user, password=password, sock=sock)
    if lst in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans":
        return []

    jobs = []
    for record in convertor.text_to_list(lst):
        jobs.append(convertor.list_to_dict(convertor.text_to_list(record, ":")))

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_job_details1(ip, port, user=None, password=None, sock=None):

    """
    This method is no more in USE
    """
    if user is None:
        lst = cl.citrusleaf_info(ip, port, "mon:cmd=getall-job", sock=sock)
    else:
        lst = cl.citrusleaf_info(ip, port, "mon:cmd=getall-job", user=user, password=password, sock=sock)
    if lst in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans" or lst == "" or lst == "error":
        return []

    jobs = {}
    keys = ["trid", "cpu", "mem", "start_time", "run_time", "disk_io_bytes", "net_io_bytes", "priority"]
    for record in lst[1:-1].split("]["):
        details = convertor.text_to_list(record, "|")
        details = dict(zip(keys, details))
        details["address"] = "%s:%s" % (ip, port)
        jobs[details["trid"]] = details

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_job_details(ip, port, user=None, password=None, sock=None):

    if user is None:
        lst = cl.citrusleaf_info(ip, port, "jobs:", sock=sock)
    else:
        lst = cl.citrusleaf_info(ip, port, "jobs:", user=user, password=password, sock=sock)
    if lst in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans" or lst == "" or lst == "error":
        return []

    jobs = {}
    for record in convertor.text_to_list(lst.strip(), ";"):
        details = convertor.list_to_dict(
            convertor.text_to_list(record, ":"), "=")
        details["address"] = "%s:%s" % (ip, port)
        jobs[details["trid"]] = details

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_index_stats(ip, port, namespace, index, user=None, password=None, sock=None):

    command = "sindex/%s/%s" % (str(namespace), str(index))
    if user is None:
        stats = cl.citrusleaf_info(ip, port, command, sock=sock)
    else:
        stats = cl.citrusleaf_info(ip, port, command, user=user, password=password, sock=sock)
    if stats in cl.ERROR_CODES:
        return InfoDict(node_status="off")
    elif stats == "ns_type=unknown":
        return InfoDict(namespace_found=False)
    elif stats == "Invalid StatsFAIL:201:NO INDEX":
        return InfoDict(index_found=False)
    stats = convertor.text_to_list(stats)
    return convertor.list_to_dict(stats)

@debugger.check_and_log
# @CachedMethod()
def get_indexes(ip, port, user=None, password=None, sock=None):

    if user is None:
        output = cl.citrusleaf_info(ip, port, "sindex", sock=sock)
    else:
        output = cl.citrusleaf_info(ip, port, "sindex", user=user, password=password, sock=sock)
    if output in cl.ERROR_CODES:
        return UNABLE_TO_CONNECT

    indexes = {}
    output = output.replace("sync_state=:", "sync_state=")
    for record in convertor.text_to_list(output):
        details = convertor.list_to_dict(convertor.text_to_list(record, ":"))
        if "ns" not in details:
            continue
        namespace = indexes.get(details["ns"], [])
        namespace.append(details)
        indexes[details["ns"]] = namespace

    return indexes

@debugger.check_and_log
def check_authenticity_of_user(ip, port, user, password, sock=None):

    response = cl.citrusleaf_info(ip, port, "service", user=user, password=password, sock=sock)
    return response


@debugger.check_and_log
# @CachedMethod()
def get_cluster_userroles(ip, port, user=None, password=None, sock=None):

    user = get_citrusleaf_cluster_user(ip, port, user=user, password=password, sock=sock)
    if user is None or user[1] is None:
        return None
    else:
        # user[1].roles is list of roles assigned to this particular user.
        return user[1].roles

@debugger.check_and_log
# @CachedMethod(60)
def get_citrusleaf_cluster_user(ip, port, user=None, password=None, sock=None):

    """
    This method will call citrusleaf.queryUser to fetch particular user roles.
    """
    if user is not None:

        # Create a citrusleaf cluster object necessary for querying Users
        citrusleaf_cluster = cl.CitrusleafCluster()

        # password field assumed to be hashed one

        citrusleaf_cluster.setUser(user, password)
        cl.cluster_setHost(citrusleaf_cluster, ip, int(port))

        user_data = cl.queryUser(citrusleaf_cluster, user)

        # Delete citrusleaf_cluster object, not needed anymore
        citrusleaf_cluster.crawler_disable = True

        del citrusleaf_cluster

        return user_data
    else:
        return None

@debugger.check_and_log
# @CachedMethod()
def is_security_enabled(ip, port, sock=None):

    """
    Assuming there is no issue with the ip and port connectivity.
    """
    try:

        output = cl.citrusleaf_info(ip, port, "service", sock=sock)
        if type(output) == str:
            return False
        if type(output) == int and output == UNABLE_TO_CONNECT:
            output = cl.citrusleaf_info(ip, port, "service", user='', password='', sock=sock)
            if type(output) == int and output != UNABLE_TO_CONNECT:
                return True
            elif output == UNABLE_TO_CONNECT:
                return UNABLE_TO_CONNECT
    except:
        return UNABLE_TO_CONNECT

@debugger.check_and_log
def create_citrusleaf_cluster(cluster_obj):

    # address = "%s:%s" % (cluster_obj._ip, cluster_obj._port)
    node_obj = cluster_obj.get_node(cluster_obj.get_seed())
    user = node_obj.user
    password = node_obj.password
    citrusleaf_cluster = cl.CitrusleafCluster()
    citrusleaf_cluster.setUser(str(user), str(password))
    cl.cluster_setHost(citrusleaf_cluster, cluster_obj._access_ip, int(cluster_obj._access_port))
    return citrusleaf_cluster
    """except:
        import sys
        sys.stderr.write('\nERROR!!')
        return 52"""


@debugger.check_and_log
def delete_citrusleaf_cluster(cluster):

    cluster.crawler_disable = True
    del cluster

@debugger.check_and_log
def add_user(cluster_obj, user, password, roles=["read"]):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.createUser(citrusleaf_cluster, user, password, roles)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def drop_user(cluster_obj, user):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.dropUser(citrusleaf_cluster, user)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def replace_roles(cluster_obj, user, roles):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.replaceRoles(citrusleaf_cluster, user, roles)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def grant_roles(cluster_obj, user, roles):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.grantRoles(citrusleaf_cluster, user, roles)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def revoke_roles(cluster_obj, user, roles):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.revokeRoles(citrusleaf_cluster, user, roles)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def set_password(cluster_obj, user, password):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.setPassword(citrusleaf_cluster, user, password)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def get_all_users(cluster_obj):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    response = cl.queryUsers(citrusleaf_cluster)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return response

@debugger.check_and_log
def change_password(cluster_obj, user, old_password, new_password):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.changePassword(citrusleaf_cluster, user, old_password, new_password)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if type(status) == int and status not in cl.ERROR_CODES:
        return True, status
    return False, status

@debugger.check_and_log
def query_role(ip, port, user, password, role):

    citrusleaf_cluster = cl.CitrusleafCluster()
    citrusleaf_cluster.setUser(str(user), str(password))
    cl.cluster_setHost(citrusleaf_cluster, ip, int(port))
    status, privileges = cl.queryRole(citrusleaf_cluster, role)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    if status != 0 :
        return status, None
    return status, privileges.privileges

@debugger.check_and_log
def query_roles(cluster_obj):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status, roles = cl.queryRoles(citrusleaf_cluster)
    if status != 0 :
        delete_citrusleaf_cluster(citrusleaf_cluster)
        return status, None
    response = []
    temp_dict = {}
    for role in roles:
        response.append({ str(role.role[0]) : role.privileges })
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return status, response

@debugger.check_and_log
def create_role(cluster_obj, role, privileges=None):

    if privileges is None:
        return -1
    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.createRole(citrusleaf_cluster, role, privileges)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return status

@debugger.check_and_log
def delete_role(cluster_obj, role):

    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.deleteRole(citrusleaf_cluster, role)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return status

@debugger.check_and_log
def add_privileges(cluster_obj, role, privileges=None):

    if privileges is None:
        return -1
    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.addPrivileges(citrusleaf_cluster, role, privileges)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return status

@debugger.check_and_log
def drop_privileges(cluster_obj, role, privileges=None):

    if privileges is None:
        return -1
    citrusleaf_cluster = create_citrusleaf_cluster(cluster_obj)
    status = cl.deletePrivileges(citrusleaf_cluster, role, privileges)
    delete_citrusleaf_cluster(citrusleaf_cluster)
    return status

@debugger.check_and_log
def fire_cmd(ip, port, cmd, user=None, password=None, sock=None):

    if user is not None:
        response = cl.citrusleaf_info(ip, port, str(cmd), user=user, password=password, sock=sock)
    else:
        response = cl.citrusleaf_info(ip, port, str(cmd), sock=sock)
    try:
        resp_list = convertor.text_to_list(response)
    except:
        return response
    return resp_list


@debugger.check_and_log
def get_node_availabilty(ip, port, **kwargs):
    info = cl.citrusleaf_info(ip, port, **kwargs)
    if info in cl.ERROR_CODES:
        return 'not_available'
    elif info == {'security error - not authenticated': ''} or info == {
                'ERROR:80:not authenticated': ''}:
        return 'secured'
    elif 'build' in info.keys():
        return 'available'
    else:
        return 'not_available'


@debugger.check_and_log
def get_dc_config(ip, xdr_port, **kwargs):
    """ Get XDR shipping configuration of a given node
    :param ip: <IP>
    :param xdr_port: <XDR-PORT>
    :return:
        {
            <dc-name>: {
                'Nodes': [{
                    'ip': <ip>, 'port': <port>, 'int_ip': <int_ip>,
                    'ext_ip': <ext_ip>
                }, ]
                'seednode': {
                    'ip': <ip>, 'port': <port>, 'int_ip': <int_ip>,
                    'ext_ip': <ext_ip>
                },
                'namespaces': [<namespace1>, ]
            }
        }
    """
    # Few AS versions work only with 'get-dc-config:'
    # and few only with 'get-dc-config' while some with both
    kwargs['names'] = 'get-dc-config:'
    output = cl.citrusleaf_info(ip, xdr_port, **kwargs)
    if output in cl.ERROR_CODES or not output:
        kwargs['names'] = 'get-dc-config'
        output = cl.citrusleaf_info(ip, xdr_port, **kwargs)
        if output in cl.ERROR_CODES or not output:
            return {}

    # Convert raw string received into key-value store
    try:
        dc_config = filter(
            lambda x: x['namespaces'],  # Only add if dc is shipping namespaces
            list(  # list & dict comprehension syntax doesn't work in python 2.6
                dict(
                    (
                        key, None if value == '' else  # nested ternary operator
                        value.split(',') if key == 'namespaces' else
                        [ip.replace('+', ':') for ip in value.split(",")]
                        if key == 'Nodes' else
                        [ip_map.split('+') for ip_map in value.split(",")]
                        if key == 'int-ext-ipmap' else value
                     )
                    for key, value in (dc.split("=") for dc in dcs.split(":"))
                )
                for dcs in output.split(";") if dcs
            )
        )
    except:  # In case format is changed; this shouldn't break AMC
        dc_config = {}

    # Mark entry against DC Name to avoid duplicates which will be
    # received from other nodes of cluster
    dc_config = dict((x.pop('DC_Name'), x) for x in dc_config)

    # Map node addresses to their external ip
    # Choose a seed node
    dc_to_remove = []
    for dc_name, dc in dc_config.iteritems():

        # if `Nodes` is empty, add it to a list so we remove it later
        if dc.get('Nodes') is None:
            dc_to_remove.append(dc_name)
            continue

        for index, node in enumerate(dc['Nodes']):
            node_ip, node_port = node.split(":")

            # Change Nodes structure to accomodate int-ext-ipmap
            dc['Nodes'][index] = {
                'ip': node_ip, 'port': node_port, 'int_ip': None, 'ext_ip': None
            }

            if dc['int-ext-ipmap']:
                for ipmap in dc['int-ext-ipmap']:
                    if node_ip in ipmap:
                        int_ip, ext_ip = ipmap

                        if node_ip == ext_ip:
                            # ip should point to internal ip
                            dc['Nodes'][index]['ip'] = int_ip

                        node_address = get_self_address(ext_ip, node_port)
                        if node_address not in cl.ERROR_CODES:
                            # Update with the configured node address
                            ip, port = node_address.split(':')
                            dc['Nodes'][index] = {
                                'ip': ip, 'port': port, 'int_ip': ip,
                                'ext_ip': ext_ip
                            }

                            # Set seed node
                            if 'seednode' not in dc:
                                dc['seednode'] = dc['Nodes'][index]
                        else:
                            # Secured node might not provide output
                            # Just store int-ext-ipmap
                            dc['Nodes'][index]['int_ip'] = int_ip
                            dc['Nodes'][index]['ext_ip'] = ext_ip

        if 'seednode' not in dc:
            dc['seednode'] = dc['Nodes'][0]  # Assign first node as seednode

        del dc['int-ext-ipmap']  # Not required anymore

    # remove DCs with no seeds
    for dc_name in dc_to_remove:
        del dc_config[dc_name]

    return dc_config


@debugger.check_and_log
def get_cluster_nodes_alternate(ip, port, **kwargs):
    """ Get alternate-address of other nodes in cluster """
    cluster_nodes = cl.citrusleaf_info(ip, port, 'services-alternate', **kwargs)
    if cluster_nodes not in cl.ERROR_CODES and cluster_nodes:
        return convertor.text_to_list(cluster_nodes)
    return []


@debugger.check_and_log
# @CachedMethod(60)
def get_node_features(ip, port, **kwargs):
    """ Get features of node """
    features = cl.citrusleaf_info(ip, port, 'features', **kwargs)
    if features not in cl.ERROR_CODES and features:
        return convertor.text_to_list(features)
    return []
